library(rootSolve)
library(utils)
library(optimx)

input_data = read.csv("input_data.csv", header = TRUE, sep = ",", dec = ".")  # read csv file 

#Functions
conc <- function (H,G0,H0,K) return((H^2+H*(G0-H0+1/K)-(H0/K)))

#Setup
bs.cycles <- 1000
initialguess.cycles <- 100
t1<- Sys.time()
pb   <- txtProgressBar(1, 100, style=3)
IGResult.tab <- matrix(, nrow=initialguess.cycles, ncol=8)
BSResult.tab <- matrix(, nrow=bs.cycles, ncol=8)

#Data
H0.data <- input_data [,1]
G0.data<- input_data [,2]
dAr1_obs.data<-input_data [,3]
dAr2_obs.data<-input_data [,4]
dAlk_obs.data<-input_data [,5]

#Setup
Max_H <- H0.data[1]
H0.bsdata<-vector(,length(H0.data))
G0.bsdata<-vector(,length(H0.data))
dAr1_obs.bsdata<-vector(,length(H0.data))
dAr2_obs.bsdata<-vector(,length(H0.data))
dAlk_obs.bsdata<-vector(,length(H0.data))

#bootstrap start
for (k in 1:bs.cycles){
  bs.dataset <- sample(1:length(H0.data),length(H0.data),replace=T)
  for (l in 1:length(H0.data)) {
    H0.bsdata[l]<-H0.data[bs.dataset[l]]
    G0.bsdata[l]<-G0.data[bs.dataset[l]]
    dAr1_obs.bsdata[l]<-dAr1_obs.data[bs.dataset[l]]
    dAr2_obs.bsdata[l]<-dAr2_obs.data[bs.dataset[l]]
    dAlk_obs.bsdata[l]<-dAlk_obs.data[bs.dataset[l]]
  }

#initial guess start  
  for (j in 1:initialguess.cycles) {
    repeat {
      K.init<- runif(1,0,10)
      dAr1_H.init<-runif(1,6.63,6.645)
      dAr1_HG.init<-runif(1,5.8,6.6)
      dAr2_H.init<-runif(1,6.48,6.49)
      dAr2_HG.init<-runif(1,5.6,6.5)
      dAlk_H.init<-runif(1,1.20,1.22)
      dAlk_HG.init<-runif(1,0.8,1.2)
      
      InitialParameters<-c(K.init,dAr1_H.init,dAr1_HG.init, dAr2_H.init, dAr2_HG.init, dAlk_H.init,dAlk_HG.init)    
 
      cH <- vector(length=length(H0.bsdata))
      for (i in 1:length(H0.bsdata)) {
        roots <-uniroot.all(f = conc, lower= 0, upper=Max_H,G0=G0.bsdata[i], H0=H0.bsdata[i],K=K.init)
        cH[i] <- min(roots)
      }

            if ((('Inf' %in% cH == FALSE)&(0.0000 %in% cH == FALSE))){
      break
      } 
    }
 
    InitialParameters<-c(K.init,dAr1_H.init,dAr1_HG.init, dAr2_H.init, dAr2_HG.init, dAlk_H.init,dAlk_HG.init)
      
#Function for optimization/minimization
    tst <- function (param,H0.min,G0.min,dAr1_obs.min, dAr2_obs.min, dAlk_obs.min) {
    for (i in 1:length(H0.min)) {
        roots <-uniroot.all(f = conc, lower=0, upper=Max_H,G0=G0.min[i],H0=H0.min[i],K=param[1])
        cH[i] <- min(roots)
      }
      cHG <- H0.min-cH
      cG <- G0.min-cHG
      dAr1_calc <- param[2]*cH/H0.min+param[3]*cHG/H0.min
      dAr2_calc <- param[4]*cH/H0.min+param[5]*cHG/H0.min
      dAlk_calc <- param[6]*cH/H0.min+param[7]*cHG/H0.min

      dif_Ar1 <- (dAr1_calc-dAr1_obs.min)
      dif_Ar2 <- (dAr2_calc-dAr2_obs.min)
      dif_Alk <- (dAlk_calc-dAlk_obs.min)

      sqdiff <- (dif_Ar1)^2+(dif_Ar2)^2+(dif_Alk)^2
      return(sum(sqdiff))
    }
#Optimize & Summarize Data
    result <- optimx(par=InitialParameters,fn=tst, method="nlminb", gr=NULL, hess=NULL, lower=0, upper=Inf, itnmax=NULL, hessian=FALSE, control=list(), H0.min=H0.bsdata,G0.min=G0.bsdata,dAr1_obs.min=dAr1_obs.bsdata,dAr2_obs.min=dAr2_obs.bsdata,dAlk_obs.min=dAlk_obs.bsdata)
    IGResult.tab[j,] <- c(result[1,1],result[1,2],result[1,3],result[1,4],result[1,5],result[1,6],result[1,7],result[1,8])
  }
#Select best (lowest sum-square-diff) set of parameters from the multiple initial guesses in each bootstrap cycle and write to file
  which.min(IGResult.tab[,8])
  BSResult.tab[k,]<-IGResult.tab[which.min(IGResult.tab[,8]),]
  write.csv(BSResult.tab, file = "1_to_1_output.csv", row.names=FALSE)
  prog <- k/bs.cycles*100
  setTxtProgressBar(pb, prog)
}

#Report final set of bootstrap data and write file
BSResult.tab
t2 <- Sys.time()
time <- t2-t1
time
write.csv(BSResult.tab, file = "1_to_1_final_output.csv", row.names=FALSE)
which.min(BSResult.tab[,8])
BSResult.tab[which.min(BSResult.tab[,8]),]


#Visualize the data for best set of parameters
K1<-BSResult.tab[which.min(BSResult.tab[,8]),1]
dAr1_H<-BSResult.tab[which.min(BSResult.tab[,8]),2]
dAr1_HG<-BSResult.tab[which.min(BSResult.tab[,8]),3]
dAr2_H<-BSResult.tab[which.min(BSResult.tab[,8]),4]
dAr2_HG<-BSResult.tab[which.min(BSResult.tab[,8]),5]
dAlk_H<-BSResult.tab[which.min(BSResult.tab[,8]),6]
dAlk_HG<-BSResult.tab[which.min(BSResult.tab[,8]),7]

cH.data <- vector(length=length(H0.data))
for (i in 1:length(H0.data)) {
cH.data[i] <- uniroot.all(f = conc, lower=0, upper=Max_H,H0=H0.data[i],G0=G0.data[i],K=K1)
}
cHG.data <- H0.data-cH.data
cG.data <- G0.data-cHG.data
dAr1_calc.data <- dAr1_H*cH.data/H0.data+dAr1_HG*cHG.data/H0.data
dAr2_calc.data <- dAr2_H*cH.data/H0.data+dAr2_HG*cHG.data/H0.data
dAlk_calc.data <- dAlk_H*cH.data/H0.data+dAlk_HG*cHG.data/H0.data
dif_Ar1.data <- (dAr1_calc.data-dAr1_obs.data)
dif_Ar2.data <- (dAr2_calc.data-dAr2_obs.data)
dif_Alk.data <- (dAlk_calc.data-dAlk_obs.data)
plot(G0.data/H0.data,dAr1_obs.data)
lines(G0.data/H0.data, dAr1_calc.data)
plot(G0.data/H0.data,dif_Ar1.data)
plot(G0.data/H0.data,dAr2_obs.data)
lines(G0.data/H0.data, dAr2_calc.data)
plot(G0.data/H0.data,dif_Ar2.data)
plot(G0.data/H0.data,dAlk_obs.data)
lines(G0.data/H0.data, dAlk_calc.data)
plot(G0.data/H0.data,dif_Alk.data)
plot(G0.data/H0.data,cH.data/H0.data, type="l", ylim=c(0,1))
lines(G0.data/H0.data,cHG.data/H0.data)
lines(G0.data/H0.data,cG.data/H0.data)
