library(rootSolve)
library(utils)
library(optimx)
fit.model <- 'weighted-shift' #(all, shift, weighted-shift)

#Model
#H + G --> HG  K11
#HG + H --> HHG K21
#HHG + H --> HHHG K31
#Functions
concH <- function (H, K11, K21, K31, H0, G0) return((K11*K21*K31)*H^4+(K11*K21+3*K11*K21*K31*G0-K11*K21*K31*H0)*H^3+(-K11*K21*H0+2*K11*K21*G0+K11)*H^2+(1+K11*G0-K11*H0)*H-H0)

#Cycles
bs.cycles <- 100
initialguess.cycles <- 20

#Time reporting and Status
t1<- Sys.time()
pb   <- txtProgressBar(1, 100, style=3)

#Empty matricies for reporting
IGResult.tab <- matrix(, nrow=initialguess.cycles, ncol=8)
BSResult.tab <- matrix(, nrow=bs.cycles, ncol=8)

#Data
input_data = read.csv("input_data.csv", header = TRUE, sep = ",", dec = ".")  # read csv file 
H0.data<-input_data [,1]
G0.data<-input_data [,2]
d1_obs.data<-input_data [,3]
d2_obs.data<-input_data [,4]
I1_obs.data<-input_data [,5]
I2_obs.data<-1-I1_obs.data
#weighting functions
d1_W <- (1/(max(d1_obs.data)-min(d1_obs.data)))^2
d2_W <- (1/(max(d2_obs.data)-min(d2_obs.data)))^2
I1_W <- (1/(max(I1_obs.data)-min(I1_obs.data)))^2
I2_W <- (1/(max(I2_obs.data)-min(I2_obs.data)))^2
#Sets maximum concentration for root solver (equal to initial H0 concentration)
Max_H <- 1.1*max(H0.data)

#Setup empty vectors for bootstrap data
H0.bsdata<-vector(,length(H0.data))
G0.bsdata<-vector(,length(H0.data))
d1_obs.bsdata<-vector(,length(H0.data))
d2_obs.bsdata<-vector(,length(H0.data))
I1_obs.bsdata<-vector(,length(H0.data))
I2_obs.bsdata<-vector(,length(H0.data))

#Start bootstrap
for (k in 1:bs.cycles){
 bs.dataset <- sample(1:length(H0.data),length(H0.data),replace=T)
  for (l in 1:length(H0.data)) {
    H0.bsdata[l]<-H0.data[bs.dataset[l]]
    G0.bsdata[l]<-G0.data[bs.dataset[l]]
    d1_obs.bsdata[l]<-d1_obs.data[bs.dataset[l]]
    d2_obs.bsdata[l]<-d2_obs.data[bs.dataset[l]]
    I1_obs.bsdata[l]<-I1_obs.data[bs.dataset[l]]
    I2_obs.bsdata[l]<-I2_obs.data[bs.dataset[l]]
  }
  for (j in 1:initialguess.cycles) {
#repeat loop that cycles until initial conditions that are solvable with uniroot.all function are chosen
    repeat {
      K11.init<- runif(1,0,1000)
      K21.init<-runif(1,0,1000)
      K31.init<-runif(1,0,1000)
      dH.init<-runif(1,6.635,6.645)
      dHG.init<-runif(1,4.5,6.5)
      dHHG.init<-runif(1,4.5,6.5)
      dHHHG.init<-runif(1,4.5,6.5) 
          
      cH <- vector(length=length(H0.bsdata))
      for (i in 1:length(H0.bsdata)) {
        roots <-uniroot.all(f = concH,K11=K11.init,K21=K21.init,K31=K31.init,H0=H0.bsdata[i],G0=G0.bsdata[i], lower= 0, upper=Max_H)
        cH[i] <- min(roots)
      }
      InitialParameters<-c(K11.init,K21.init,K31.init,dH.init,dHG.init,dHHG.init,dHHHG.init)

      if ((('Inf' %in% cH == FALSE)&(0.0000 %in% cH == FALSE))){
      break
      } 
      
    }
  
    #Paramters: K11,K21,K31,dH,dHG,dHHG,dHHHG
    InitialParameters<-c(K11.init,K21.init,K31.init,dH.init,dHG.init,dHHG.init,dHHHG.init)
    LowerParameterLimit<-c(0,0,0,0,0,0,0)
    UpperParameterLimit<-c(Inf,Inf,Inf,Inf,Inf,Inf,Inf)

    #Function for minimization
    tst <- function (param,H0,G0,d1_obs,d2_obs,I1_obs,I2_obs) {
      K11<-param[1]
      K21<-param[2]
      K31<-param[3]
      dH<-param[4]
      dHG<-param[5]
      dHHG<-param[6]
      dHHHG<-param[7]
      cH <- vector(length=length(H0))
      one <- rep(1,length(H0))
      for (i in 1:length(H0)) {
        roots <-uniroot.all(f = concH,K11=K11,K21=K21,K31=K31,H0=H0[i],G0=G0[i], lower= 0, upper=Max_H)
        cH[i]<-min(roots)
      }
      cG <- G0/(one+K11*cH+K11*K21*cH^2+K11*K21*K31*cH^3)
      cHG <- K11*cH*cG
      cHHG <- K21*cHG*cH
      cHHHG <- K31*cHHG*cH
      d1_calc <- dH*cH/(cH+cHG)+dHG*cHG/(cH+cHG)
      d2_calc <- 2*dHHG*cHHG/(2*cHHG+3*cHHHG)+3*dHHHG*cHHHG/(2*cHHG+3*cHHHG)
      I1_calc <- (cH+cHG)/H0
      I2_calc <- (2*cHHG+3*cHHHG)/H0
      differ_d1 <- (d1_calc-d1_obs)
      differ_d2 <- (d2_calc-d2_obs)
      differ_I1 <- (I1_calc-I1_obs)
      differ_I2 <- (I2_calc-I2_obs)

      if (fit.model=='all') {
      sqdiff <- (differ_d1)^2+(differ_d2)^2+(differ_I1)^2+(differ_I2)^2
      } 
      if (fit.model=='shift') {
      sqdiff <- (differ_d1)^2+(differ_d2)^2
      }
      if (fit.model=='weighted-shift') {
      sqdiff <- d1_W*(differ_d1)^2+d2_W*(differ_d2)^2+I1_W*(differ_I1)^2+I2_W*(differ_I2)^2
      }
      
      return(sum(sqdiff))
    }
    #Optimize
    result <- optimx(par=InitialParameters,tst, method=("nlminb"), gr=NULL, hess=NULL, lower=LowerParameterLimit, upper=UpperParameterLimit, itnmax=NULL, hessian=FALSE, control=list(), H0=H0.bsdata,G0=G0.bsdata,d1_obs=d1_obs.bsdata,d2_obs=d2_obs.bsdata, I1_obs=I1_obs.bsdata, I2_obs=I2_obs.bsdata)
    IGResult.tab[j,] <- c(result[1,1],result[1,2],result[1,3],result[1,4],result[1,5],result[1,6],result[1,7],result[1,8])
  }
  which.min(IGResult.tab[,8])
  BSResult.tab[k,]<-IGResult.tab[which.min(IGResult.tab[,8]),]
  write.csv(BSResult.tab, file = "3_to_1_H_HG_output.csv", row.names=FALSE)
  prog <- k/bs.cycles*100
  setTxtProgressBar(pb, prog)
}
BSResult.tab
t2 <- Sys.time()
time <- t2-t1
time
write.csv(BSResult.tab, file = "3_to_1_H_HG_final_output.csv", row.names=FALSE)
which.min(BSResult.tab[,8])      
BSResult.tab[which.min(BSResult.tab[,8]),]      
      
#Visualize
K11<-BSResult.tab[which.min(BSResult.tab[,8]),1]
K21<-BSResult.tab[which.min(BSResult.tab[,8]),2]
K31<-BSResult.tab[which.min(BSResult.tab[,8]),3]
dH<-BSResult.tab[which.min(BSResult.tab[,8]),4]
dHG<-BSResult.tab[which.min(BSResult.tab[,8]),5]
dHHG<-BSResult.tab[which.min(BSResult.tab[,8]),6]
dHHHG<-BSResult.tab[which.min(BSResult.tab[,8]),7]

cH <- vector(length=length(H0.data))
one <- rep(1,length(H0.data))
for (i in 1:length(H0.data)) {
  roots <-uniroot.all(f = concH,K11=K11,K21=K21,K31=K31,H0=H0.data[i],G0=G0.data[i], lower= 0, upper=Max_H)
  cH[i]<-min(roots)
}
cG <- G0.data/(one+K11*cH+K11*K21*cH^2+K11*K21*K31*cH^3)
cHG <- K11*cH*cG
cHHG <- K21*cHG*cH
cHHHG <- K31*cHHG*cH
d1_calc <- dH*cH/(cH+cHG)+dHG*cHG/(cH+cHG)
d2_calc <- 2*dHHG*cHHG/(2*cHHG+3*cHHHG)+3*dHHHG*cHHHG/(2*cHHG+3*cHHHG)
I1_calc <- (cH+cHG)/H0.data
I2_calc <- (2*cHHG+3*cHHHG)/H0.data
differ_d1 <- (d1_calc-d1_obs.data)
differ_d2 <- (d2_calc-d2_obs.data)
differ_I1 <- (I1_calc-I1_obs.data)
differ_I2 <- (I2_calc-I2_obs.data)

plot(G0.data/H0.data,d1_obs.data)
lines(G0.data/H0.data, d1_calc)
plot(G0.data/H0.data,differ_d1) 
plot(G0.data/H0.data,d2_obs.data)
lines(G0.data/H0.data, d2_calc)
plot(G0.data/H0.data,differ_d2)
plot(G0.data/H0.data,I1_obs.data)
lines(G0.data/H0.data, I1_calc)
plot(G0.data/H0.data,differ_I1)
plot(c(0,G0.data/H0.data),c(1,cH/H0.data), type="l", ylim=c(0,1))
lines(c(0,G0.data/H0.data),c(0,cHG/H0.data))
lines(c(0,G0.data/H0.data),c(0,cHHG/H0.data))
lines(c(0,G0.data/H0.data),c(0,cHHHG/H0.data))
lines(c(0,G0.data/H0.data),c(0,cG/H0.data))  
